#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

from upk2esphome.generator import invert
from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def generate(yr: YamlResult, config: dict, opts: Opts):
    keys = ["netled", "netled1", "wfst"]

    found = False
    for key in keys:
        led_pin = config.get(f"{key}_pin")
        led_inv = config.get(f"{key}_pin", None) == 0
        if led_pin is None:
            continue
        if found:
            yr.warn("Multiple netled pins found!")
            continue
        found = True

        yr.log(f"Status LED: pin P{led_pin}, inverted {led_inv}")
        status_led = {
            "pin": f"P{led_pin}",
        }
        invert(status_led, led_inv)
        yr.data["status_led"] = status_led



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

from upk2esphome.generator import invert
from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def generate(yr: YamlResult, config: dict, opts: Opts):
    keys = [
        "bt1_pin",
        "bt2_pin",
        "bt3_pin",
        "bt4_pin",
        "total_bt_pin",
        "rl1_pin",
        "rl2_pin",
        "rl3_pin",
        "rl4_pin",
        "rl_type",
    ]
    if not any(key in config for key in keys):
        return
    yr.log("Switch/plug config")

    switches = []
    for i in range(0, 10):
        rl_pin = config.get(f"rl{i}_pin", None)
        rl_inv = config.get(f"rl{i}_lv", None) == 0
        led_pin = config.get(f"led{i}_pin", None)
        led_inv = config.get(f"led{i}_lv", None) == 0
        bt_pin = config.get(f"bt{i}_pin", None)
        bt_inv = config.get(f"bt{i}_lv", None) == 0
        if rl_pin is None:
            continue

        yr.log(f" - relay {i}: pin P{rl_pin}")
        yr.found = True
        switch = {
            "platform": "gpio",
            "id": f"switch_{i}",
            "name": f"Relay {i}",
            "pin": f"P{rl_pin}",
        }
        invert(switch, rl_inv)
        if led_pin is not None:
            yr.log(f" - LED {i}: pin P{led_pin}")
            yr.found = True
            output = {
                "platform": "ledc",
                "id": f"output_led_{i}",
                "pin": f"P{led_pin}",
            }
            invert(output, led_inv)
            yr.output(output)
            light = {
                "platform": "monochromatic",
                "id": f"light_switch_{i}",
                "output": output["id"],
            }
            yr.light(light)
            switch["on_turn_on"] = [
                {"light.turn_on": light["id"]},
            ]
            switch["on_turn_off"] = [
                {"light.turn_off": light["id"]},
            ]
        if bt_pin is not None:
            yr.log(f" - button {i}: pin P{bt_pin}")
            yr.found = True
            binary = {
                "platform": "gpio",
                "id": f"binary_switch_{i}",
                "pin": f"P{bt_pin}",
                "on_press": {
                    "then": [
                        {"switch.toggle": switch["id"]},
                    ],
                },
            }
            invert(binary, bt_inv)
            yr.binary(binary)
        switches.append(switch["id"])
        yr.switch(switch)

    bt_pin = config.get(f"total_bt_pin", None)
    bt_inv = config.get(f"total_bt_lv", None) == 0
    if bt_pin is not None:
        yr.log(f" - all-toggle button: pin P{bt_pin}")
        yr.found = True
        binary = {
            "platform": "gpio",
            "id": f"binary_switch_all",
            "pin": f"P{bt_pin}",
            "on_press": {"then": []},
        }
        invert(binary, bt_inv)
        for switch in switches:
            binary["on_press"]["then"].append({"switch.toggle": switch})
        yr.binary(binary)



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

import json
import re
from typing import Union


def parse_user_param_key(value: str) -> dict | None:
    value = re.sub(r"([^{}\[\]:,]+)", r'"\1"', value)
    value = re.sub(r'"([1-9][0-9]*|0)"', r"\1", value)
    value = re.sub(",}", "}", value)
    value = json.loads(value)
    return value


def parse_input(data: Union[str, dict]) -> dict:
    if isinstance(data, str):
        data = data.strip()
    loaders = [
        json.loads,
        parse_user_param_key,
        dict,
    ]

    value: dict | None = None
    for loader in loaders:
        try:
            value = loader(data)
            break
        except Exception:
            pass

    if not isinstance(value, dict):
        raise ValueError("Input format unrecognized")

    if "device_configuration" in value:
        value = value["device_configuration"]
    elif "schemas" in value:
        raise ValueError(
            "The specified Cloudcutter JSON doesn't contain device configuration"
        )

    if "user_param_key" in value:
        value = value["user_param_key"]
    elif "gw_bi" in value:
        raise ValueError(
            "The specified storage JSON doesn't contain device configuration"
        )

    if "Jsonver" in value or "jv" in value or "crc" in value:
        value = dict(sorted(value.items()))
        return value

    raise ValueError("JSON value found, but the config is missing")



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def generate(yr: YamlResult, config: dict, opts: Opts):
    if opts.common:
        yr.data["logger"] = {}
    if opts.web_server:
        yr.data["web_server"] = {}
        yr.data["captive_portal"] = {}
        yr.data["mdns"] = {}
    if opts.common:
        yr.data["api"] = {"password": opts.api_password}
        yr.data["ota"] = {"password": opts.ota_password}
        yr.data["wifi"] = {
            "ssid": opts.wifi_ssid,
            "password": opts.wifi_password,
            "ap": {},
        }
    if opts.restart:
        yr.button(
            {
                "platform": "restart",
                "name": "Restart",
            }
        )
        yr.data["debug"] = {
            "update_interval": "30s",
        }
        yr.text_sensor(
            {
                "platform": "debug",
                "reset_reason": {"name": "Reset Reason"},
            }
        )
    if opts.uptime:
        yr.sensor(
            {
                "platform": "uptime",
                "name": "Uptime",
            }
        )
    if opts.lt_version:
        yr.text_sensor(
            {
                "platform": "libretiny",
                "version": {"name": "LibreTiny Version"},
            }
        )



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

import re

import yaml

from upk2esphome.input import parse_input
from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def invert(data: dict, inverted: bool, key: str = "pin"):
    if inverted:
        data[key] = {
            "number": data[key],
            "inverted": True,
        }


def generate_yaml(config: str, opts: Opts) -> YamlResult:
    from .parts import bulb, module, monitor, netled, static, switch

    parts = [module, static, bulb, switch, netled, monitor]
    yr = YamlResult()

    try:
        config = parse_input(config)
    except ValueError as e:
        yr.error(str(e))
        config = {}

    for part in parts:
        part.generate(yr, config, opts)

    if not yr.found and not yr.errors:
        yr.error(
            "No actual components were added! This means that the type of your device "
            "is not yet supported by this program. This includes for example "
            "thermometers, water leak sensors, or fan controllers."
        )

    yr.text = yaml.dump(yr.data, sort_keys=False)
    yr.text = re.sub(r"'!secret ([\w_]+)'", r"!secret \1", yr.text)
    yr.text = re.sub(r"\n([a-z])", r"\n\n\1", yr.text)
    yr.text = yr.text.replace("'", '"')
    yr.text = yr.text.replace(" {}", "")
    return yr



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

from upk2esphome.generator import invert
from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def generate(yr: YamlResult, config: dict, opts: Opts):
    keys = [
        "chip_type",
        "over_cur",
        "over_vol",
        "lose_vol",
        "ele_pin",
        "vi_pin",
        "sel_pin_pin",
        "sel_pin_lv",
        "sample_resistor",
        "work_voltage",
        "vol_def",
    ]
    if not any(key in config for key in keys):
        return
    if "chip_type" not in config:
        yr.warn("Found some power monitoring keys, but no chip_type!")
        return

    resistor = config.get("sample_resistor", None)
    resistor = config.get("resistor", resistor)
    chip_type = config["chip_type"]
    match chip_type:
        case 0 | 1:  # BL0937 |  HLW8012
            ele_pin = config.get("ele_pin", None)
            vi_pin = config.get("vi_pin", None)
            sel_pin = config.get("sel_pin_pin", None)
            sel_inv = config.get("sel_pin_lv", None) == 1
            chip = "BL0937" if chip_type == 0 else "HLW8012"
            yr.log(
                f"Power monitoring chip {chip}: "
                f"CF=P{ele_pin}, CF1=P{vi_pin}, SEL={sel_pin}"
            )
            yr.found = True
            sensor = {
                "platform": "hlw8012",
                "model": chip,
                "cf_pin": f"P{ele_pin}",
                "cf1_pin": f"P{vi_pin}",
                "sel_pin": f"P{sel_pin}",
                "current": {"name": f"{chip} Current"},
                "voltage": {"name": f"{chip} Voltage"},
                "power": {"name": f"{chip} Power"},
                "energy": {"name": f"{chip} Energy"},
            }
            if resistor is not None:
                yr.log(f" - shunt resistor: {resistor} mΩ")
                sensor["current_resistor"] = f"{resistor/1000:.03f} ohm"
            invert(sensor, sel_inv, "sel_pin")
            yr.sensor(sensor)

        case 2:  # HLW8032
            yr.warn("HLW8032 is not supported")

        case 3:  # BL0942
            rx_pin = config.get("ele_rx", None)
            tx_pin = config.get("ele_tx", None)
            rx_pin = f"P{rx_pin}" if rx_pin else "RX1"
            tx_pin = f"P{tx_pin}" if tx_pin else "TX1"
            chip = "BL0942"
            yr.log(f"Power monitoring chip {chip}: " f"RX={rx_pin}, TX={tx_pin}")
            yr.found = True
            uart = {
                "id": "uart_bus",
                "tx_pin": tx_pin,
                "rx_pin": rx_pin,
                "baud_rate": 4800,
                "stop_bits": 1,
            }
            yr.data["uart"] = uart
            sensor = {
                "platform": "bl0942",
                "uart_id": uart["id"],
                "current": {"name": f"{chip} Current"},
                "voltage": {"name": f"{chip} Voltage"},
                "power": {"name": f"{chip} Power", "filters": {"multiply": -1}},
                "energy": {"name": f"{chip} Energy"},
                "frequency": {"name": f"{chip} Frequency", "accuracy_decimals": 2},
            }
            yr.sensor(sensor)

        case _:
            yr.warn(f"Unrecognized power monitoring chip {chip_type}")



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.


class YamlResult:
    data: dict
    text: str
    logs: list[str]
    warnings: list[str]
    errors: list[str]
    found: bool

    def __init__(self):
        self.data = {}
        self.text = ""
        self.logs = []
        self.warnings = []
        self.errors = []
        self.found = False

    def log(self, text: str):
        self.logs.append(text)

    def warn(self, text: str):
        self.warnings.append(text)

    def error(self, text: str):
        self.errors.append(text)

    def component(self, data: dict):
        if "external_components" not in self.data:
            self.data["external_components"] = []
        self.data["external_components"].append(data)

    def output(self, data: dict):
        if "output" not in self.data:
            self.data["output"] = []
        self.data["output"].append(data)

    def light(self, data: dict):
        if "light" not in self.data:
            self.data["light"] = []
        self.data["light"].append(data)

    def switch(self, data: dict):
        if "switch" not in self.data:
            self.data["switch"] = []
        self.data["switch"].append(data)

    def binary(self, data: dict):
        if "binary_sensor" not in self.data:
            self.data["binary_sensor"] = []
        self.data["binary_sensor"].append(data)

    def sensor(self, data: dict):
        if "sensor" not in self.data:
            self.data["sensor"] = []
        self.data["sensor"].append(data)

    def text_sensor(self, data: dict):
        if "text_sensor" not in self.data:
            self.data["text_sensor"] = []
        self.data["text_sensor"].append(data)

    def button(self, data: dict):
        if "button" not in self.data:
            self.data["button"] = []
        self.data["button"].append(data)



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

from .generator import generate_yaml
from .input import parse_input
from .opts import Opts
from .result import YamlResult

__all__ = [
    "parse_input",
    "generate_yaml",
    "YamlResult",
    "Opts",
]



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

from upk2esphome.generator import invert
from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult

colors = ["red", "green", "blue", "cold", "warm"]
platforms = {
    "rgb": ("rgb", {}, {}),
    "rgbw": ("rgbw", {"warm": "white"}, {}),
    "rgbc": ("rgbw", {"cold": "white"}, {}),
    "rgbcw": (
        "rgbww",
        {"cold": "cold_white", "warm": "warm_white"},
        {
            "color_interlock": True,
            "cold_white_color_temperature": "6500 K",
            "warm_white_color_temperature": "2700 K",
        },
    ),
    "cw": (
        "cwww",
        {"cold": "cold_white", "warm": "warm_white"},
        {
            "cold_white_color_temperature": "6500 K",
            "warm_white_color_temperature": "2700 K",
        },
    ),
    "c": ("monochromatic", {"cold": "output"}, {}),
    "w": ("monochromatic", {"warm": "output"}, {}),
}
cmods = {
    "rgbcw": "rgbww",
    "rgb": "rgb",
    "cw": "cwww",
    "c": "monochromatic",
    "rgbc": "rgbw",
}


def get_platform(found: set[str]) -> str | None:
    key = sorted(c[0] for c in found)
    for name, platform in platforms.items():
        if sorted(name) == key:
            return platform
    return None


def gen_pwm(yr: YamlResult, config: dict) -> set[str]:
    found = set()
    for color in colors:
        pin = config.get(f"{color[0]}_pin", None)
        inv = config.get(f"{color[0]}_lv", None) == 0
        if pin is None:
            continue
        found.add(color)

        yr.log(f" - color {color}: pin P{pin}, inverted {inv}")
        output = {
            "platform": "ledc",
            "id": f"output_{color}",
            "pin": f"P{pin}",
        }
        invert(output, inv)
        yr.output(output)
    return found


def gen_i2c_sm2235(yr: YamlResult, config: dict):
    scl = config.get("iicscl", None)
    sda = config.get("iicsda", None)
    if not scl or not sda:
        yr.warn("I2C pins not found")
        return None

    yr.data["sm2235"] = {
        "clock_pin": f"P{scl}",
        "data_pin": f"P{sda}",
    }

    cur_color = config.get("2235ccur", None)
    cur_white = config.get("2235wcur", None)
    if cur_color is not None:
        yr.data["sm2235"]["max_power_color_channels"] = cur_color
        yr.log(f" - color channels power: {cur_color}")
    if cur_white is not None:
        yr.data["sm2235"]["max_power_white_channels"] = cur_white
        yr.log(f" - white channels power: {cur_white}")

    found = set()
    for color in colors:
        channel = f"iic{color[0]}"
        if channel not in config:
            continue
        found.add(color)

        yr.log(f" - color {color}: channel {config[channel]}")
        output = {
            "platform": "sm2235",
            "id": f"output_{color}",
            "channel": config[channel],
        }
        yr.output(output)

    return found


def gen_i2c_sm2135eh(yr: YamlResult, config: dict):
    scl = config.get("iicscl", None)
    sda = config.get("iicsda", None)
    if not scl or not sda:
        yr.warn("I2C pins not found")
        return None

    yr.component(
        {
            "source": "github://pr#3850",
            "components": ["sm2135"],
            "refresh": "10 min",
        }
    )
    yr.data["sm2135"] = {
        "clock_pin": f"P{scl}",
        "data_pin": f"P{sda}",
    }

    cur_color = config.get("ehccur", config.get("iicccur", None))
    cur_white = config.get("ehwcur", config.get("iicwcur", None))
    if cur_color is not None:
        cur_color = 10 + cur_color * 5
    else:
        cur_color = config.get("campere", None)
    if cur_white is not None:
        cur_white = 10 + cur_white * 5
    else:
        cur_white = config.get("wampere", None)

    if cur_color is not None:
        yr.data["sm2135"]["rgb_current"] = f"{cur_color}mA"
        yr.log(f" - color channels current: {cur_color} mA")
    if cur_white is not None:
        yr.data["sm2135"]["cw_current"] = f"{cur_white}mA"
        yr.log(f" - white channels current: {cur_white} mA")

    found = set()
    for color in colors:
        channel = f"iic{color[0]}"
        if channel not in config:
            continue
        found.add(color)

        yr.log(f" - color {color}: channel {config[channel]}")
        output = {
            "platform": "sm2135",
            "id": f"output_{color}",
            "channel": config[channel],
        }
        yr.output(output)

    return found


def gen_i2c_bp5758d(yr: YamlResult, config: dict):
    scl = config.get("iicscl", None)
    sda = config.get("iicsda", None)
    if not scl or not sda:
        yr.warn("I2C pins not found")
        return None

    yr.data["bp5758d"] = {
        "clock_pin": f"P{scl}",
        "data_pin": f"P{sda}",
    }

    cur_rgb = config.get("drgbcur", None)
    cur_cold = config.get("dccur", None)
    cur_warm = config.get("dwcur", None)

    found = set()
    for color in colors:
        channel = f"iic{color[0]}"
        if channel not in config:
            continue
        found.add(color)

        current = (
            cur_cold if color == "cold" else cur_warm if color == "warm" else cur_rgb
        )

        yr.log(f" - color {color}: channel {config[channel]}, current {current} mA")
        output = {
            "platform": "bp5758d",
            "id": f"output_{color}",
            "channel": config[channel],
        }
        if current:
            output["current"] = current
        yr.output(output)

    return found


def gen_i2c_bp1658cj(yr: YamlResult, config: dict):
    scl = config.get("iicscl", None)
    sda = config.get("iicsda", None)
    if not scl or not sda:
        yr.warn("I2C pins not found")
        return None

    yr.data["bp1658cj"] = {
        "clock_pin": f"P{scl}",
        "data_pin": f"P{sda}",
    }

    cur_color = config.get("cjccur", None)
    cur_white = config.get("cjwcur", None)
    if cur_color is not None:
        yr.data["bp1658cj"]["max_power_color_channels"] = cur_color
        yr.log(f" - color channels power: {cur_color}")
    if cur_white is not None:
        yr.data["bp1658cj"]["max_power_white_channels"] = cur_white
        yr.log(f" - white channels power: {cur_white}")

    found = set()
    for color in colors:
        channel = f"iic{color[0]}"
        if channel not in config:
            continue
        found.add(color)

        yr.log(f" - color {color}: channel {config[channel]}")
        output = {
            "platform": "bp1658cj",
            "id": f"output_{color}",
            "channel": config[channel],
        }
        yr.output(output)

    return found


def generate(yr: YamlResult, config: dict, opts: Opts):
    # remove duplicate channels
    if config.get("iicc", None) == config.get("iicw", None):
        config.pop("iicw", None)

    chips = {
        "PWM": (gen_pwm, ["r_pin", "g_pin", "b_pin", "c_pin", "w_pin"]),
        "SM2235": (gen_i2c_sm2235, ["2235ccur", "2235wcur"]),
        "SM2135EH": (
            gen_i2c_sm2135eh,
            ["ehccur", "ehwcur", "wampere", "campere", "iicccur", "iicwcur"],
        ),
        "BP5758D": (gen_i2c_bp5758d, ["dccur", "dwcur", "drgbcur"]),
        "BP1658CJ": (gen_i2c_bp1658cj, ["cjccur", "cjwcur"]),
    }
    for name, (func, keys) in chips.items():
        if not any(key in config for key in keys):
            continue
        yr.log(f"{name} bulb")
        yr.found = True
        # find colors
        found = func(yr, config)
        if not found:
            continue
        # find platform for colors
        platform = get_platform(found)
        if not platform:
            yr.warn(f"Unknown light platform for colors: {found}")
            continue
        name, mapping, opts = platform
        # build light component
        light = {
            "platform": name,
            "id": f"light_{name}",
            "name": "Light",
            **opts,
        }
        # add channels
        for color in colors:
            if color not in found:
                continue
            key = mapping.get(color, color)
            light[key] = f"output_{color}"
        # add light component
        yr.light(light)
        # check cmod matching
        if "cmod" in config:
            cmod = config["cmod"]
            if cmods[cmod] != name:
                yr.warn(f"Module 'cmod': {cmod} doesn't match platform {name}")



#  Copyright (c) Kuba Szczodrzyński 2023-4-22.

from dataclasses import dataclass


@dataclass
class Opts:
    esphome_block: bool = True
    name_mac: bool = True
    common: bool = True
    web_server: bool = True
    restart: bool = True
    uptime: bool = True
    lt_version: bool = True
    wifi_ssid: str = ""
    wifi_password: str = ""
    ota_password: str = ""
    api_password: str = ""



#  Copyright (c) Kuba Szczodrzyński 2023-4-22.

if __name__ == "__main__":
    import sys
    from glob import glob
    from os.path import dirname, join

    from .generator import generate_yaml
    from .opts import Opts

    for file in glob(join(dirname(__file__), "tests", "*.txt")):
        if len(sys.argv) == 2 and sys.argv[1] not in file:
            continue
        with open(file, "r") as f:
            d = f.read().strip()
        print(file)
        yr = generate_yaml(d, Opts())
        print("\n".join(yr.logs))
        print("\n".join(yr.warnings))
        print("\n".join(yr.errors))
        print(yr.text)
        print("-" * 80)



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def generate(yr: YamlResult, config: dict, opts: Opts):
    if "module" in config:
        module = config["module"]
    else:
        yr.warn("No module type found! You have to set the board: manually")
        if opts.esphome_block:
            yr.data["esphome"] = {
                "name": "UPK2ESPHome",
            }
            if opts.name_mac:
                yr.data["esphome"]["name_add_mac_suffix"] = True
            yr.data["libretiny"] = {
                "board": "FIX_ME",
                "framework": {
                    "version": "dev",
                },
            }
        return

    match module[0:2].upper():
        case "WB":
            board = "generic-bk7231t-qfn32-tuya"
        case "CB":
            board = "generic-bk7231n-qfn32-tuya"
        case _:
            yr.warn("Only BK7231T and BK7231N chips are currently supported")
            return

    yr.log(f"Found {module} config!")

    if opts.esphome_block:
        yr.data["esphome"] = {
            "name": config["module"],
        }
        if opts.name_mac:
            yr.data["esphome"]["name_add_mac_suffix"] = True
        yr.data["libretiny"] = {
            "board": board,
            "framework": {
                "version": "dev",
            },
        }



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

from js import jQuery

from upk2esphome import Opts, generate_yaml


def on_run_click():
    opts = {}
    options = jQuery("#options input")
    for i in range(options.length):
        el = options[i]
        el_id = el.id
        if not el_id.startswith("opts_"):
            continue
        el_value = el.value if el.type == "text" else el.checked
        opts[el_id[5:]] = el_value
    opts = Opts(**opts)

    config = jQuery("#input").val()
    yr = generate_yaml(config, opts)
    jQuery("#output").val(yr.text)
    jQuery("#logs").html("<pre>" + "<br>".join(yr.logs) + "</pre>")
    jQuery("#warnings").html(
        "".join(
            f'<div class="alert alert-danger" role="alert">{e}</div>' for e in yr.errors
        )
        + "".join(
            f'<div class="alert alert-warning" role="alert">{w}</div>'
            for w in yr.warnings
        )
    )
